"""
preprocessing and shorthand functions for dataframes, mappings, and ndarrays.
"""
from typing import Any

import numpy as np
import pandas as pd


def itemize_numpy(obj: Any):
    """
    convert objects of numpy dtypes to python scalars. in this context,
    primarily for json serialization.
    """
    if isinstance(obj, np.generic):
        return obj.item()
    return obj


def dupe_df_block(dataframe: pd.DataFrame, repeats: int):
    """
    create a new dataframe by duplicating "dataframe" "repeats" times
    """
    return pd.DataFrame(
        np.repeat(dataframe.values, repeats, axis=0),
        columns=dataframe.columns,
    )


# TODO, maybe: allow this to work with columns that have collections
#  in individual cells
def check_and_drop_duplicate_columns(dataframe: pd.DataFrame):
    """
    drop duplicate columns -- in both name and value -- from a dataframe.
    if equal-named columns are _not_ equal-valued, raises a ValueError.
    """
    extra_columns = dataframe.columns[dataframe.columns.duplicated()]
    if len(extra_columns) == 0:
        return dataframe
    for column in extra_columns:
        test_equality = (dataframe.loc[:, column] == dataframe.loc[:, column])
        if not test_equality.all(axis=None):
            raise ValueError
    return dataframe.loc[:, ~dataframe.columns.duplicated()]


def extract_constants(df, to_dict=True, drop_constants=False, how='rows'):
    """
    extract 'constant' values from a dataframe -- by default, columns with
    the same value in each row; if how == 'columns', then indices with the
    same value in each column. if to_dict is True, transforms them into a
    dictionary with keys equal to column/index names and values equal to the
    constant values. if drop_constants is True, also removes these
    constant-valued rows/columns from the returned dataframe.
    returns the constants (as a dict if specified) and either the
    variable-only or full original dataframe (as specified)
    """
    if how == 'rows':
        axis = 0
    elif how == 'columns':
        axis = 1
    else:
        raise ValueError(f"unknown how {how}")
    constant_indices = df.nunique(axis=axis) == 1
    if axis == 0:
        constants = df.loc[:, constant_indices]
        variables = df.loc[:, ~constant_indices]
    else:
        constants = df.loc[constant_indices]
        variables = df.loc[~constant_indices]
    if to_dict:
        constants = constants.iloc[0].to_dict()
    if drop_constants:
        return constants, variables
    return constants, df


def split_on(
    df: pd.DataFrame, predicate: pd.Series
) -> [pd.DataFrame, pd.DataFrame]:
    return df.loc[predicate], df.loc[~predicate]


def pdstr(str_method_name, *str_args, **str_kwargs):
    """
    creates a mappable function that accesses .str methods of passed Series
    """

    def replacer(series: pd.Series):
        method = getattr(series.str, str_method_name)
        return method(*str_args, **str_kwargs)

    return replacer



"""
generic text-parsing functions designed for quickly yoinking metadata
from labels, particularly on networked and/or huge filesystems, without
sophisticated parsing. probably also useful for other things. sometimes
appropriate, sometimes not.
"""
from ast import literal_eval
from functools import cache
from io import BytesIO
import os
from pathlib import Path
import re
from typing import Callable, Iterable, Mapping, Optional, Union


def make_scraper(label_text: str) -> Callable:
    """
    makes a little function that scrapes a particular text for patterns
    """

    def scrape(pattern):
        result = re.search(pattern, label_text)
        if result is None:
            return None
        try:
            return literal_eval(result.group(1))
        except (ValueError, SyntaxError):
            return result.group(1)

    return scrape


def get_label_text(label: Union[Path, str], byte_size: int = 35592) -> str:
    """
    why not overload the label-getter? literally no reason
    """
    if isinstance(label, str):
        label = Path(label)
    with open(label, "rb") as file:
        return file.read(byte_size).decode()


# special-case block regex
SUBFRAME_GROUP = re.compile(r"SUBFRAME.*?END", re.DOTALL)


def scrape_subframe(label_text: str) -> tuple:
    """
    special case -- making a sequence from multiple pvl fields
    in a block. can generalize if necessary. output is:
    first line (row), first line sample (column),
    total lines, total line samples
    if image is not subframed, should be (1, 1, 1200, 1648)
    """
    return literal_eval(
        ",".join(
            re.findall(r"\d+", re.search(SUBFRAME_GROUP, label_text).group())
        )
    )


# cached versions for faster operation on networked filesystems.
cached_label_loader = cache(get_label_text)


@cache
def scrape_from_file(label, pattern):
    return make_scraper(cached_label_loader(label))(pattern)


# heuristic for max label size. we know it's not a real rule.
MAX_LABEL_SIZE = 500 * 1024


def head_file(fn_or_reader, nbytes):
    head_buffer = BytesIO()
    if not hasattr(fn_or_reader, "read"):
        fn_or_reader = open(fn_or_reader, 'rb')
    head_buffer.write(fn_or_reader.read(nbytes))
    fn_or_reader.close()
    head_buffer.seek(0)
    return head_buffer


KNOWN_LABEL_ENDINGS = (
    b'END\r\n',  # PVL
    b'\x00{3}',  # just null bytes
)


def trim_label(fn, max_size = MAX_LABEL_SIZE, raise_for_failure=False):
    head = head_file(fn, max_size).read()
    # TODO: add some logging or whatever i guess
    for ending in KNOWN_LABEL_ENDINGS:
        if (endmatch := re.search(ending, head)) is not None:
            return head[:endmatch.span()[1]].decode()
    if raise_for_failure:
        raise ValueError("couldn't find a label ending")
    return head.decode()


cached_label_trimmer = cache(trim_label)


def scrape_patterns(
    label: Union[Path, str],
    metadata_regex: Mapping,
    supplemental_search_function: Optional[Callable] = None,
    label_loader = cached_label_loader
) -> dict:
    """
    grabs all the patterns you specify from whatever.
    """
    label_text = label_loader(label)
    # little closure for neatness
    scrape = make_scraper(label_text)
    # general-case fields
    metadata = {
        field: scrape(pattern) for field, pattern in metadata_regex.items()
    }
    # optional insertion point for special cases that cannot be defined with
    # naive regex, default values, etc.
    if supplemental_search_function is not None:
        metadata |= supplemental_search_function(label, label_text)
    return metadata


def bulk_scrape_metadata(
    files: Iterable, pattern_scraper: Callable
) -> list[dict]:
    """
    scrapes metadata from all the files you pass it, and that's that
    """
    metaframe = []
    for file in files:
        metaframe.append(pattern_scraper(file))
    return metaframe


# cached filesystem functions for execution speed on networked filesystems
cached_ls = cache(os.listdir)
cached_exists = cache(os.path.exists)



"""structured data and manipulators thereof"""
from collections import defaultdict
from copy import copy
from functools import reduce, partial
from operator import methodcaller, add, getitem, eq
from typing import (
    Mapping, Collection, Any, Union, Sequence, MutableMapping, Callable, Type
)

from cytoolz import merge

from dustgoggles.func import naturals, is_it, splat


def to_records(nested: Mapping, accumulated_levels=None, level_names=None):
    level_names = naturals() if level_names is None else iter(level_names)
    records = []
    accumulated_levels = (
        {} if accumulated_levels is None else accumulated_levels
    )
    level_name = next(level_names)
    for category, mapping in nested.items():
        if all([isinstance(value, Mapping) for value in mapping.values()]):
            branch = accumulated_levels.copy()
            branch[level_name] = category
            records += to_records(mapping, branch, copy(level_names))
        else:
            category_dict = accumulated_levels | {level_name: category}
            flat = mapping | category_dict
            records.append(flat)

    return records


def unnest(mapping_mapping):
    unnested = []
    for category, mapping in mapping_mapping.items():
        unnested.append(
            {
                str(category) + "_" + str(key): value
                for key, value in mapping.items()
            }
        )
    return merge(unnested)


class NestingDict(defaultdict):
    """
    shorthand for automatically-nesting dictionary -- i.e.,
    insert a series of keys at any depth into a NestingDict
    and it automatically creates all needed levels above.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_factory = NestingDict

    __repr__ = dict.__repr__

    def todict(self):
        """warning: this function may permanently modify lower levels"""
        # note: could be made depth-first by using methodcaller('todict')
        # rather than dict
        return dict(
            dig_and_edit(
                self, lambda _, v: is_it(NestingDict)(v), lambda _, v: dict(v)
            )
        )


def listify(thing):
    """Always a list, for things that want lists"""
    if isinstance(thing, Collection):
        if not isinstance(thing, str):
            return list(thing)
    return [thing]


def reindex_mapping(mapping: Mapping) -> Mapping[int, Any]:
    assert isinstance(mapping, Mapping), "reindex_mapping only takes Mappings"
    return {
        ix: value for ix, value in zip(range(len(mapping)), mapping.values())
    }


def enumerate_as_mapping(sequence):
    if sequence is None:
        return {}
    if isinstance(sequence, Mapping):
        return sequence
    return {ix: element for ix, element in enumerate(sequence)}


def insert_after(
    new_key: Any, new_value: Any, prior_key: Any, mapping: Mapping
) -> Mapping:
    new_dict = {}
    for key, value in mapping.items():
        new_dict[key] = value
        if key == prior_key:
            new_dict[new_key] = new_value
    return new_dict


def get_from_all(key, mappings, default=None):
    """
    get all values of "key" from each dict or whatever in "mappings"
    """
    if isinstance(mappings, Mapping):
        view = mappings.values()
    else:
        view = mappings
    return list(map(methodcaller("get", key, default), view))


def getitemattr(collection: Union[Sequence, Mapping], key: Any) -> Any:
    """
    getter that attempts both getattr and getitem (intended
    for named tuples nested inside of dicts, etc)
    """
    try:
        return getitem(collection, key)
    except (KeyError, IndexError, TypeError):
        return getattr(collection, key)


def get_from(collection, keys, default=None):
    """
    hierarchical (toolz-style) extension of get_itemattr() --
    (hierarchical list of keys, collection ->
    item of collection, possibly from a nested collection)
    """
    level = collection
    for key in keys:
        try:
            level = getitemattr(level, key)
        except AttributeError:
            return default
    return level


def constant_to_dig_predicate(putative_predicate, match="key", default=eq):
    if isinstance(putative_predicate, Callable):
        return putative_predicate
    if match == "key":
        return keyonly(partial(default, putative_predicate))
    elif match == "value":
        return valonly(partial(default, putative_predicate))
    elif match == "item":
        return partial(default, putative_predicate)


def keyonly(func):
    def onlykey(key, _value):
        return func(key)
    return onlykey


def valonly(func):
    def onlyval(_key, value):
        return func(value)
    return onlyval


def _evaluate_diglevel(mapping, predicate, mtypes):
    # note: this is a little awkward and is done to circumvent the fact
    # that checking isinstance against the ABCs is relatively slow
    # but we would like to support many types of Mapping
    # that are not dreamt of in this library.
    nests = [v for v in mapping.values() if isinstance(v, mtypes)]
    level_items = [(k, v) for k, v in mapping.items() if predicate(k, v)]
    return level_items, nests


def dig_for_all(mapping, predicate, mtypes):
    level_items, nests = _evaluate_diglevel(mapping, predicate, mtypes)
    if nests:
        level_items += reduce(
            add, [dig_for_all(nest, predicate, mtypes) for nest in nests]
        )
    return level_items


def dig_all_wrapper(mapping, ref, match, base_pred, element_ix, mtypes):
    predicate = constant_to_dig_predicate(ref, match, base_pred)
    items = dig_for_all(mapping, predicate, mtypes)
    if len(items) == 0:
        return None
    if element_ix is None:
        return items
    return [item[element_ix] for item in items]


def dig_for_values(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_all_wrapper(mapping, ref, match, base_pred, 1, mtypes)


def dig_for_keys(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_all_wrapper(mapping, ref, match, base_pred, 0, mtypes)


def dig_for_items(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_all_wrapper(mapping, ref, match, base_pred, None, mtypes)


def dig_for(mapping, predicate, mtypes):
    """
    return the first item, using a breadth-first search, matching ref
    """
    level_items, nests = _evaluate_diglevel(mapping, predicate, mtypes)
    if level_items:
        return level_items
    if not nests:
        return None
    dug, iternests = None, iter(nests)
    while dug is None:
        try:
            dug = dig_for(next(iternests), predicate, mtypes)
        except StopIteration:
            return None
    return dug


def dig_wrapper(mapping, ref, match, base_pred, element_ix, mtypes):
    predicate = constant_to_dig_predicate(ref, match, base_pred)
    item = dig_for(mapping, predicate, mtypes)
    if item is None:
        return None
    if element_ix is None:
        return item[0]
    return item[0][element_ix]


def dig_for_value(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_wrapper(mapping, ref, match, base_pred, 1, mtypes)


def dig_for_key(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_wrapper(mapping, ref, match, base_pred, 0, mtypes)


def dig_for_item(mapping, ref, match="key", base_pred=eq, mtypes=(dict,)):
    return dig_wrapper(mapping, ref, match, base_pred, None, mtypes)


def dig_and_edit(
    mapping: MutableMapping,
    filter_func: Callable[[Any, Any], Any],
    setter_func: Callable[[Any, Any], Any],
    mtypes: tuple[Type[MutableMapping]] = (dict,)
) -> MutableMapping:
    matches = tuple(filter(splat(filter_func), mapping.items()))
    for key, value in matches:
        mapping[key] = setter_func(key, value)
    for nest in [v for v in mapping.values() if isinstance(v, mtypes)]:
        dig_and_edit(nest, filter_func, setter_func, mtypes)
    return mapping


# TODO: turn pivot.split_on into a dispatch function in structures,
#  replace downstream
def separate_by(collection, ref):
    hits = []
    misses = []
    for item in collection:
        if ref(item) is True:
            hits.append(item)
        else:
            misses.append(item)
    return hits, misses



import ast
import json
import pickle

from multiprocessing.shared_memory import SharedMemory
from typing import Any, Callable, Optional

from dustgoggles.codex.memutilz import (
    open_block, fetch_block_bytes, exists_block
)
from dustgoggles.func import zero


def json_codec_factory() -> tuple[Callable, Callable]:
    """
    generate a pair of encoder / decoder functions that can be inserted into
    the memorize / remember steps of a mnemonic or used on their own.

    this is a very simple codec that simply uses JSON serialization and
    deserialization. only objects that are stable through JSON serialization
    can/should be passed to this codec.
    """
    def encode(value: Any) -> bytes:
        """return bytestring of json serialization of `value`"""
        return json.dumps(value).encode()

    def decode(blob: bytes) -> Any:
        """return json deserialization of `blob`"""
        return json.loads(blob)
    return encode, decode


def ast_codec_factory():
    """
    generate a pair of encoder / decoder functions that can be inserted into
    the memorize / remember steps of a mnemonic or used on their own.

    this is a very simple codec that uses the Python abstract syntax tree
    serializer and deserializer. It will reliably serialize a wider variety of
    Python literals than using JSON serialization, but is much slower.
    """
    def encode(value: Any) -> bytes:
        return str(value).encode()

    def decode(blob: bytes) -> Any:
        return ast.literal_eval(blob.decode())
    return encode, decode


def pickle_codec_factory():
    """
    generate a pair of encoder / decoder functions that can be inserted into
    the memorize / remember steps of a mnemonic or used on their own.

    this just pickles things. It is a useful generic codec that tends to
    produce behaviors that mirror default Python interprocess communication.
    """
    def encode(value: Any) -> bytes:
        return pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL)

    def decode(blob: bytes) -> Any:
        return pickle.loads(blob)

    return encode, decode


def generic_mnemonic_factory():
    """
    generate a pair of save / load functions that can be used to send and
    retrieve objects from shared memory. this is a general-purpose mnemonic
    that will work for many sorts of objects. these save / load functions can
    be partially evaluated to create bound methods (see objects in
    codex.implements). alternatively, different encoders, decoders, address
    schemes, etc. can be dynamically passed into these functions at runtime.
    """

    def memorize(
        value: Any,
        address: str,
        exists_ok: bool,
        encode: Callable[[Any], bytes]
    ) -> str:
        """
        serialize `value` using `encode` and place it into a shared memory
        block at `address`. If a block already exists at `address` and
        exists_ok is not True, will raise a FileExistsError. returns
        `address`.
        """
        if (exists_ok is False) and exists_block(address):
            raise FileExistsError
        encoded = encode(value)
        size = len(encoded)
        block = open_block(address, size, True, True)
        block.buf[:] = encoded
        return address

    def remember(
        address: str,
        fetch: bool = True,
        decode: Callable[[bytes], Any] = zero
    ) -> Any:
        """
        open the shared memory block at `address`. if `fetch` is False,
        return the block itself. otherwise, copy the bytes into memory,
        decode them with `decode`, and return the result.
        """
        if fetch is False:
            return SharedMemory(name=address)
        return decode(fetch_block_bytes(address))

    return memorize, remember


def numpy_mnemonic_factory() -> tuple[Callable, Callable]:
    """
    generate a pair of save / load functions that can be used to send and
    retrieve numpy ndarrays to and from shared memory. they can be partially
    evaluated to work as bound methods of higher-level objects, or used
    as standalone objects. unlike generic_mnemonic_factory, these functions
    do not have swappable codecs; they use the array's own buffer protocol.

    TODO: consider implementing inline codecs that pass the buffer /
     memoryview rather than the binary blob in order to support compressing
     / sparsifying / etc. arrays.
    """
    import numpy as np

    def memorize_array(
        array: np.ndarray, address: str, exists_ok: bool, keep_open: bool
    ) -> tuple[dict, Optional[SharedMemory]]:
        """
        generate an array backed by a shared memory block at `address` and
        fill it with the contents of `array`. return a dict containing
        metadata necessary to fetch and reconstruct the array.
        """
        block = open_block(address, array.size * array.itemsize, exists_ok)
        shared_array = np.ndarray(
            array.shape, dtype=array.dtype, buffer=block.buf
        )
        shared_array[:] = array[:]
        metadata = {
            "name": block.name,
            "dtype": array.dtype.str,
            "shape": array.shape,
            "size": array.size * array.itemsize
        }
        if keep_open is False:
            return metadata, None
        return metadata, block

    def remember_array(
        metadata_json: str, fetch: bool = True, copy: bool = True
    ):
        metadata = json.loads(metadata_json)
        block = SharedMemory(name=metadata["name"])
        if fetch is False:
            return block
        array = np.ndarray(
            metadata['shape'], dtype=metadata['dtype'], buffer=block.buf
        )
        if copy is True:
            return array.copy(), None
        return array, block

    return memorize_array, remember_array



"""
high-level abstractions for working with Python shared memory.

notes
----
## indexes and locks: performance
Like any synchronization objects, shared indices and locks introduce some
overhead. The overhead is negligible for applications that access these
objects somewhat infrequently, but can be significant for applications that
access them at high rates, especially from many threads, as processes spinlock
waiting for access to shared resources.

If you are certain that your operations are intrinsically thread-safe (perhaps
processes have some other synchronization strategy), or your application is
tolerant of failed or corrupted writes, you can deactivate locking by passing
no_lockout=True to AbstractNotepad or ShareableIndex constructors. (This will
propagate from an AbstractNotepad to its index.)

For even more performance improvements, you can pass index_type=FakeIndex to
a Notepad constructor. This will, of course, completely deactivate indexing:
you will not be able to explicitly list the keys of the TagIndex. This may be
acceptable if you have top-down control over all processes sharing this
Notepad address space or are otherwise able to a priori determine the possible
set of keys.

This WILL NOT work with GridPaper, which requires metadata stored in a
MetaTagIndex in order to correctly format the numpy arrays it retrieves from
shared memory.

The use of FakeIndex also makes automated cleanup with .close() or
cleanup_on_exit impossible, and you must either rely on the resource tracker,
explicitly unlink any leftover shared memory blocks on exit, or accept the
possibility of memory leaks.

## cleanup_on_exit / deactivate_shared_memory_resource_tracker()
If SharedMemory objects are instantiated in a top-level process,
especially in the __main__ namespace, the multiprocessing.resource_tracker
process generally does a good job of cleaning them up. However, sometimes we
need a little more manual control for the following reasons:
    1. blocks initially created in child processes may not always be cleanly
    and automatically unlinked, particularly if:
        a. they are killed at the OS level
        b. they never terminate
        c. the parent process is very long-running
        d. they raise especially gnarly exceptions
    2. on Mac and Linux, multiprocessing.resource_tracker often does too
    good a job of unlinking SharedMemory objects, making them unstable
    under the following conditions:
        a. use by processes that do not inherit a resource tracker from the
        same parent -- this can include both processes executed by separate
        parents and processes executed by parents that did not themselves
        instantiate a resource tracker for their children to inherit
        b. maybe others
    3. also on Mac and Linux, high-rate updates (> ~40/sec) to memory blocks
    can confuse the resource tracker, causing it to throw spurious warnings
    about leaked shared memory objects if everything is in fact unlinked
    cleanly -- these are usually followed by non-spurious but pointless
    warnings that it can't unlink those already-unlinked blocks. These
    warnings occur at exit and so are difficult to suppress.
For these reasons, you might simply want to deactivate the resource tracker:
see deactivate_shared_memory_resource_tracker() in dustgoggles.codex.memutilz.
If you do that, but you don't _want_ to create a memory leak -- which you
might, if you are trying to make a persistent in-memory object cache, but in
most cases you probably don't -- someone has to clean these objects up
manually.

Cleanup_on_exit is a convenient way to handle either this cleanup issue or
listed in 1.a - 1.d, but bear in mind that it still won't effectively clean up
in the event of some kinds of hard crash of the topmost process! You also
probably don't want to set it inside processes that will terminate before you
need to use shared memory objects they created again.

If you _do_ experience memory leaks due to unavoidable instability in the non-
dustgoggles portions of your code, they are fairly easy to clean up in most
Mac / Linux Python implementations, as SharedMemory objects are simply files
in /dev/shm and can be freely deleted at the OS level using rm or whatever. It
is somewhat more challenging to free leaked blocks on Windows.
"""
import atexit
import datetime as dt
import os
import time
from abc import abstractmethod, ABC
from collections import namedtuple
from functools import partial
from hashlib import sha256
from multiprocessing.shared_memory import SharedMemory
from pathlib import Path
from random import randint
from typing import Any, Union, Optional, Callable

from cytoolz import juxt

from dustgoggles.codex.codecs import (
    generic_mnemonic_factory,
    json_codec_factory,
    pickle_codec_factory,
    numpy_mnemonic_factory,
)
from dustgoggles.codex.memutilz import (
    open_block, fetch_block_bytes, delete_block, exists_block
)
from dustgoggles.func import zero


# debug helpers

def here():
    return f"{dt.datetime.now().isoformat()[-9:]},{os.getpid()}"


def log(message):
    with open("../holding/dumplog.csv", "a+") as file:
        file.write(message + "\n")


def printstack(stack):
    return ";".join(
        [
            f"{Path(frame.filename).name}:{frame.lineno}:{frame.function}"
            for frame in stack
        ]
    )


HASHER = partial(sha256, usedforsecurity=False)


class FakeLock:
    """
    shared memory synchronization that doesn't synchronize or use shared
    memory. can be used as a mock object for tests, or passed to
    ShareableIndex objects to decrease their thread safety in exchange for
    modest boosts to performance.
    """
    def __init__(self, *_, **__):
        FakeLockTuple = namedtuple("FakeLock", ["unlink", "close"])
        self.lock = FakeLockTuple(unlink=zero, close=zero)
        self.debug = False

    def __enter__(self):
        pass

    def __exit__(self, *_, **__):
        pass

    def acquire(self, *args, **kwargs):
        pass

    def release(self, *args, **kwargs):
        pass

class SlidingLock:
    """
    shared memory synchronization primitive. uses shared memory 'filesystem'
    backend as an atomic write register.
    """
    def __init__(self, name, debug=False):
        self.name = name
        self.lock = None
        self.count = open_block(f"{self.name}_count", 5)
        self.number = 0
        self.floor = None
        self.marker = None
        self.acquired = False
        self.spins = 0
        self.max_spins = None
        self.delay = 0
        self.debug = debug

    def __enter__(self):
        try:
            self._renumber()
            self.acquire()
        except Exception as ex:
            if getattr(self, "debug") is True:
                log(f"{here()},{self.number},crashing out: {ex}")
            self.release()
            raise ex
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if getattr(self, "debug") is True:
            log(f"{here()},{self.number},entering cleanup")
        self.release()

    def _renumber(self):
        self.floor = int.from_bytes(self.count.buf.tobytes(), "little")
        self.number = self.floor + 1
        while self.marker is None:
            try:
                self.marker = SharedMemory(
                    f"{self.name}_{self.number}", size=1, create=True
                )
            except FileExistsError:
                self.number += 1
        if getattr(self, "debug") is True:
            log(f"{here()},{self.number},taking slot {self.number}")

    def _check_acquirement(self):
        for other_number in reversed(range(self.floor, self.number)):
            if exists_block(f"{self.name}_{other_number}_lock"):
                if getattr(self, "debug") is True:
                    log(f"{here()},{self.number},{other_number} has precedence")
                # # # # #
                self.wait()
                return False
        if getattr(self, "debug") is True:
            log(f"{here()},{self.number},taking precedence")
        return True

    def acquire(self):
        try:
            self.lock = SharedMemory(
                f"{self.name}_{self.number}_lock", size=1, create=True
            )
            if getattr(self, "debug") is True:
                log(f"{here()},{self.number},placed lock")
        except FileExistsError:
            if getattr(self, "debug") is True:
                log(f"{here()},{self.number},already have lock")
            self.acquired = True
        while self.acquired is False:
            self.acquired = self._check_acquirement()

    def wait(self):
        self.spins += 1
        if self.max_spins is not None:
            if self.spins > self.max_spins:
                if getattr(self, "debug") is True:
                    log(f"{here()},{self.number},failure (waiting timeout)")
                raise TimeoutError("timed out waiting for lock")
        time.sleep(self.delay)

    def release(self):
        if self.acquired is True:
            self.count.buf[:] = self.number.to_bytes(5, "little")
            if getattr(self, "debug") is True:
                log(f"{here()},{self.number},set floor to {self.number}")
        for thing in ("lock", "marker"):
            block = getattr(self, thing)
            if block is None:
                continue
            block.unlink()
            block.close()
            setattr(self, thing, None)
            if getattr(self, "debug") is True:
                log(f"{here()},{self.number},released {thing}")
        self.acquired = False

    def close(self):
        for obj in self.lock, self.count, self.marker:
            try:
                delete_block(obj.name)
            except (FileNotFoundError, TypeError, AttributeError):
                continue


class ShareableIndex(ABC):
    """
    abstract base class for indexes that exist in shared memory and can be
    accessed from multiple processes. they are used to back various sorts of
    notepads in this module, but can also be used independently.
    """
    def __init__(
        self,
        name: str = None,
        create: bool = True,
        cleanup_on_exit: bool = False,
        no_lockout: bool = False,
        debug=False,
        **_
    ):
        if (create is False) and not exists_block(name):
            raise FileNotFoundError(
                f"Shared memory block {name} does not exist and "
                f"create=False passed. Construct this index with "
                f"create=True if you want to initialize a new index."
            )
        if no_lockout is True:
            self.lock_class = FakeLock
        else:
            self.lock_class = SlidingLock
        if cleanup_on_exit is True:
            atexit.register(self.close)
        self.debug = debug

    @abstractmethod
    def update(self):
        """
        retrieve the contents of this index from shared memory.
        """
        pass

    @abstractmethod
    def add(self, key, value=None):
        """
        add a key to the index. Some types of index support assigning a value
        to individual keys; some do not.
        """
        pass

    @abstractmethod
    def __delitem__(self, key):
        """delete an item from the index."""
        pass

    @abstractmethod
    def close(self):
        """destroy this object and everything about it."""
        pass

    def remove(self, key):
        """"remove a key from this index."""
        return self.__delitem__(key)


class FakeIndex(ShareableIndex):
    def __init__(self, *_, **__):
        super().__init__(no_lockout=True)

    def update(self, *_, **__):
        pass

    def add(self, *_, **__):
        pass

    def __delitem__(self, *_, **__):
        pass

    def close(self, *_, **__):
        pass

    def remove(self, *_, **__):
        pass


class TagIndex(ShareableIndex):
    """
    fast index based around predictable link structures and reverse hash maps
    in shared memory
    """
    def __init__(
        self,
        name=None,
        create=True,
        cleanup_on_exit=False,
        debug=False,
        no_lockout=False,
    ):
        """
        Args:
            name: name / address of the shared memory backing the index.
            create: create a new index if none exists. will _not_
                overwrite an existing index.
            cleanup_on_exit: delete everything about this object on
                process  exit. see notes on cleanup_on_exit in the top-level
                docstring for codex.implements.
            debug: dump debug information to a log file.
            no_lockout: don't use a lock on this index. decreases thread
                safety for modest gains in performance.
        """
        if name is None:
            name = str(randint(100000, 999999))
        self.name = name
        super().__init__(name, create, cleanup_on_exit, no_lockout, debug)
        if create is False:
            self.memory = SharedMemory(name=name)
        else:
            self.memory = SharedMemory(name=name, create=True, size=8)
        self.cache = []
        self.length = None

    def hash_address(self, key):
        if not isinstance(key, bytes):
            key = str(key).encode()
        namehash = HASHER(key).hexdigest()
        return f"{self.name}_name_{namehash}"

    def key_address(self, index_ix):
        return f"{self.name}_key_{index_ix}"

    def update_length(self):
        try:
            self.length = int.from_bytes(self.memory.buf.tobytes(), "little")
        except ValueError as error:
            if "not enough values" not in str(error):
                raise error
            return self.update_length()

    def update(self):
        self.cache = []
        for ix, key in self._enumerate_keys():
            self.cache.append(key)
        return self.cache

    def check(self, target) -> Optional[int]:
        if exists_block(self.hash_address(target)):
            nameblock = SharedMemory(self.hash_address(target))
            return int.from_bytes(nameblock.buf.tobytes(), "little")
        return None

    def _enumerate_keys(self):
        self.update_length()
        for ix in range(self.length):
            try:
                keyblock = SharedMemory(self.key_address(ix))
                key = keyblock.buf.tobytes().decode()
            except FileNotFoundError:
                continue
            yield ix, key

    def _add(self, key, exists_ok) -> Optional[int]:
        found_at = self.check(key)
        if found_at is not None:
            if exists_ok is True:
                if getattr(self, "debug") is True:
                    log(f"{here()},,{key} already in index at {found_at}")
                return found_at
            raise FileExistsError
        self.update_length()
        new_length_bytes = (self.length + 1).to_bytes(8, "little")
        self.memory.buf[:] = new_length_bytes
        keybytes = str(key).encode()
        keyblock = open_block(
            self.key_address(self.length),
            create=True,
            size=len(keybytes),
            overwrite=True
        )
        keyblock.buf[:] = keybytes
        position_bytes = self.length.to_bytes(8, "little")
        nameblock = open_block(self.hash_address(key), create=True, size=8)
        nameblock.buf[:] = position_bytes
        if getattr(self, "debug") is True:
            log(f"{here()},,set {self.length} to {key}")
        return self.length

    def add(self, key, value=None, exists_ok=True):
        if value is not None:
            raise ValueError("TagIndex does not support value-setting.")
        with self.lock_class(self.name, self.debug):
            return self._add(key, exists_ok)

    def _delitem(self, target):
        found_at = self.check(target)
        if found_at is None:
            raise KeyError(f"{target} not found in this index.")
        self.update_length()
        if found_at != self.length - 1:
            lastblock = SharedMemory(self.key_address(self.length - 1))
            lastkey = lastblock.buf.tobytes()
            moved_block = open_block(
                self.key_address(found_at),
                size=len(lastkey),
                overwrite=True
            )
            moved_block.buf[:] = lastkey
            nameblock = SharedMemory(self.hash_address(lastkey))
            nameblock.buf[:] = found_at.to_bytes(8, "little")
            lastblock.unlink()
            lastblock.close()
            if getattr(self, "debug") is True:
                log(
                    f"{here()},,overwrote block at position {found_at} "
                    f"with block at position {self.length - 1} "
                )
        else:
            delete_block(self.key_address(found_at))
            if getattr(self, "debug") is True:
                log(f"{here()},,deleted block at position {found_at}")
        delete_block(self.hash_address(target))
        new_length_bytes = (self.length - 1).to_bytes(8, "little")
        self.memory.buf[:] = new_length_bytes
        if getattr(self, "debug") is True:
            log(
                f"{here()},,deleted {target} at index slot {found_at} "
                f"and decremented index counter"
            )
        return found_at

    def __delitem__(self, target):
        with self.lock_class(self.name, self.debug):
            return self._delitem(target)

    def close(self):
        self.memory.unlink()
        self.memory.close()
        if exists_block(f"{self.name}_count"):
            block = SharedMemory(f"{self.name}_count")
            block.unlink()
            block.close()
        atexit.unregister(self.close)


class MetaTagIndex(TagIndex):
    """
    very slightly slower and slightly more complex version of TagIndex that
    permits attaching metadata to keys. reference usage is in GridPaper, which
    uses a MetaTagIndex to store metadata about dtypes and dimensions of numpy
    arrays in order to accurately and rapidly recreate them from raw bytes.
    """
    def __init__(
        self,
        name=None,
        create=True,
        cleanup_on_exit=False,
        debug=False,
        no_lockout=False,
    ):
        """
        Args:
            name: name / address of the shared memory backing the index.
            create: create a new index if none exists. will _not_ overwrite an
                existing index.
            cleanup_on_exit: delete everything about this object on process
                exit. see notes on cleanup_on_exit in the top-level docstring
                for codex.implements.
            debug: activate debug-level logging
            no_lockout: don't use a lock on this index. decreases thread
                safety for modest gains in performance.
        """
        super().__init__(name, create, cleanup_on_exit, debug, no_lockout)
        memorize, remember = generic_mnemonic_factory()
        encode, decode = json_codec_factory()
        self.memorize = partial(memorize, exists_ok=True, encode=encode)
        self.remember = partial(remember, fetch=True, decode=decode)
        if create is True:
            self.memory = SharedMemory(self.name)

    def val_address(self, index_ix):
        return f"{self.name}_val_{index_ix}"

    def _load_key(self, ix):
        keyblock = SharedMemory(self.key_address(ix))
        return keyblock.buf.tobytes().decode()

    def _load_value(self, ix):
        try:
            return self.remember(self.val_address(ix))
        except FileNotFoundError:
            return

    def _enumerate(self, get_keys=False, get_values=False):
        loaders = [
            loader if arg is True else zero
            for arg, loader
            in zip((get_keys, get_values), (self._load_key, self._load_value))
        ]
        load = juxt(loaders)
        self.update_length()
        for ix in range(self.length):
            key, val = load(ix)
            output = [ix]
            if get_keys is True:
                output.append(key)
            if get_values is True:
                output.append(val)
            yield tuple(output)

    def keys(self):
        return [key for _, key in self._enumerate(True)]

    def values(self):
        return [val for _, val in self._enumerate(False, True)]

    def items(self):
        return [(key, val) for _, key, val in self._enumerate(True, True)]

    def add(self, key, value=None, exists_ok=True):
        with self.lock_class(self.name, self.debug):
            writehead = super()._add(key, exists_ok)
            if value is not None:
                self.memorize(value, self.val_address(writehead))
            return writehead

    def __getitem__(self, key):
        found_at = self.check(key)
        if found_at is None:
            raise KeyError(f"{key} not found in index.")
        valblock = SharedMemory(self.val_address(found_at))
        return valblock.buf.tobytes()

    def __delitem__(self, target):
        with self.lock_class(self.name, self.debug):
            found_at = super().__delitem__(target)
            if (
                (found_at == self.length - 1)
                or not exists_block(self.val_address(self.length - 1))
            ):
                delete_block(self.val_address(found_at), True)
            elif exists_block(self.val_address(self.length - 1)):
                valblock = SharedMemory(self.val_address(self.length - 1))
                valbytes = valblock.buf.tobytes()
                moved_block = open_block(
                    self.val_address(found_at), size=len(valbytes), overwrite=True
                )
                moved_block.buf[:] = valbytes
                delete_block(self.val_address(self.length - 1))
        return found_at


class AbstractNotepad(ABC):
    def __init__(
        self,
        prefix=None,
        create=False,
        cleanup_on_exit=False,
        index_type=None,
        codec_factory: Optional[Callable] = None,
        mnemonic_factory: Optional[Callable] = None,
        update_on_init=True,
        debug=False,
        **index_kwargs
    ):
        self.debug = debug
        if prefix is None:
            prefix = randint(100000, 999999)
        self.prefix = prefix
        if mnemonic_factory is not None:
            self.memorize, self.remember = mnemonic_factory()
        if codec_factory is not None:
            if mnemonic_factory is None:
                raise TypeError("can't specify codec without a mnemonic")
            encode, decode = codec_factory()
            self.memorize = partial(self.memorize, encode=encode)
            self.remember = partial(self.remember, decode=decode)
        if index_type is None:
            raise TypeError("must specify an index type")
        try:
            self.index = index_type(
                f"{prefix}_index",
                create,
                cleanup_on_exit,
                debug,
                **index_kwargs
            )
            if update_on_init is True:
                self.update_index()
        except FileNotFoundError:
            raise FileNotFoundError(
                "the index space for this object has not been initialized "
                "(or has been deleted) and create=False was passed. Try "
                "constructing this object with create=True."
            )
        if cleanup_on_exit is True:
            atexit.register(self.close)

    def get_raw(self, key):
        return fetch_block_bytes(self.address(key))

    def address(self, key):
        return f"{self.prefix}_{key}"

    def update_index(self) -> Union[dict, tuple]:
        return self.index.update()

    @abstractmethod
    def __getitem__(self, key):
        pass

    @abstractmethod
    def get(self, key, default, fetch):
        pass

    @abstractmethod
    def __setitem__(self, key, value, exists_ok: bool = True):
        pass

    @abstractmethod
    def set(self, key, value, exists_ok: bool = True):
        pass

    def keys(self):
        return self.update_index()

    @abstractmethod
    def values(self):
        pass

    @abstractmethod
    def items(self):
        pass

    def iterkeys(self):
        for key in self.update_index():
            yield key

    def itervalues(self):
        """return an iterator over entries in the cache"""
        for key in self.update_index():
            yield self[key]

    def iteritems(self):
        """return an iterator over key / value pairs in the cache"""
        for key in self.update_index():
            yield key, self[key]

    @abstractmethod
    def __delitem__(self, key):
        pass

    @abstractmethod
    def dump(self, key, fn):
        pass

    @abstractmethod
    def close(self, dump):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def __str__(self):
        pass

    def __repr__(self):
        return self.__str__()


class Notepad(AbstractNotepad):
    """generic read-write cache"""
    def __init__(
        self,
        prefix=None,
        create=False,
        cleanup_on_exit=False,
        index_type=TagIndex,
        codec_factory=pickle_codec_factory,
        mnemonic_factory=generic_mnemonic_factory,
        update_on_init=False,
        debug=False,
        **index_kwargs
    ):
        super().__init__(
            prefix,
            create,
            cleanup_on_exit,
            index_type,
            codec_factory,
            mnemonic_factory,
            update_on_init,
            debug,
            **index_kwargs
        )

    def __getitem__(self, key, fetch=True):
        result = self.remember(self.address(key), fetch)
        if result is None:
            raise KeyError
        return result

    def get(self, key, default=None, fetch=True):
        try:
            return self.__getitem__(key, fetch)
        except KeyError:
            return default

    def __setitem__(self, key, value, meta=None, exists_ok: bool = True):
        if isinstance(key, str) and key.startswith("index"):
            raise KeyError("'index' is a reserved key prefix")
        try:
            self.index.add(key, meta, exists_ok=exists_ok)
            self.memorize(value, self.address(key), exists_ok)
            if self.debug is True:
                log(f"{here()},,set key={key} to {value}")
        except FileExistsError:
            raise KeyError(
                f"{key} already exists in this object's cache. pass "
                f"exists_ok=True to overwrite it."
            )

    def set(self, key: str, value: Any, meta=None, exists_ok: bool = True):
        return self.__setitem__(key, value, meta, exists_ok)

    def __delitem__(self, key):
        if key in (["index", "index_lock"]):
            raise KeyError("'index' and 'index_lock' are reserved key names")
        try:
            delete_block(self.address(key))
            self.index.remove(key)
        except FileNotFoundError:
            raise KeyError(f"{key} not apparently assigned")

    def values(self):
        # TODO: something with MemoryViews?
        raise NotImplementedError("Try using the itervalues method for now")

    def items(self):
        # TODO: something with MemoryViews?
        raise NotImplementedError("Try using the iteritems method for now")

    def dump(self, key, fn=None, mode="wb"):
        if fn is None:
            fn = f"{self.prefix}_{key}".replace(".", "_")
        with open(fn, mode) as file:
            # noinspection PyTypeChecker
            file.write(self.get_raw(key))

    def close(self, dump=False):
        try:
            keys = self.update_index().copy()
            for key in keys:
                if dump is True:
                    self.dump(key)
                del self[key]
            self.index.close()
        except AttributeError:
            pass
        atexit.unregister(self.close)

    def clear(self):
        for key in self.update_index():
            del self[key]

    def __str__(self):
        return f"{self.__class__.__name__} with keys {self.update_index()}"

    def __repr__(self):
        return self.__str__()


class GridPaper(AbstractNotepad):

    def __init__(
        self,
        prefix=None,
        create=False,
        cleanup_on_exit=False,
        index_type=MetaTagIndex,
        update_on_init=False,
        debug=False,
        **index_kwargs
    ):
        # noinspection PyTypeChecker
        super().__init__(
            prefix,
            create,
            cleanup_on_exit,
            index_type,
            mnemonic_factory=numpy_mnemonic_factory,
            update_on_init=update_on_init,
            debug=debug,
            **index_kwargs
        )
        self._open_blocks = []

    def __getitem__(self, key, fetch=True, copy=True):
        try:
            result, block = self.remember(self.index[key], fetch, copy)
            if copy is False:
                self._open_blocks.append(block)
        except FileNotFoundError:
            raise KeyError(f"{key} not found in this object's space.")
        return result

    def get(self, key, default=None, fetch=True, copy=False):
        try:
            return self.__getitem__(key, fetch, copy)
        except KeyError:
            return default

    def __setitem__(
        self, key, value, exists_ok: bool = True, keep_open: bool = False
    ):
        if isinstance(key, str) and key.startswith("index"):
            raise KeyError("'index' is a reserved namespace.")
        try:
            metadata, block = self.memorize(
                value, self.address(key), exists_ok, keep_open
            )
            if keep_open is True:
                self._open_blocks.append(block)
        except FileExistsError:
            raise KeyError(
                f"{key} already exists in this object's cache. pass "
                f"exists_ok=True to overwrite it."
            )
        self.index.add(key, metadata)

    def _add_index_key(self, key):
        raise TypeError

    def set(self, key, value, exists_ok: bool = True):
        return self.__setitem__(key, value, exists_ok)

    def values(self):
        # TODO: something with MemoryViews?
        raise NotImplementedError("Try using the itervalues method for now")

    def items(self):
        # TODO: something with MemoryViews?
        raise NotImplementedError("Try using the iteritems method for now")

    def dump(self, key, fn=None):
        if fn is None:
            fn = f"{self.prefix}_{key}".replace(".", "_")
        self[key].tofile(fn)

    def __delitem__(self, key):
        if key in (["index", "index_lock"]):
            raise KeyError("'index' and 'index_lock' are reserved key names")
        try:
            delete_block(self.address(key))
            self.index.remove(key)
            for block in self._open_blocks:
                if block.name == key:
                    block.close()
                    block.unlink()
        except FileNotFoundError:
            raise KeyError(f"{key} not apparently assigned")

    def close(self, dump=False):
        for key in self.update_index():
            if dump is True:
                self.dump(key)
            del self[key]
        self.index.close()
        atexit.unregister(self.close)

    def clear(self):
        for key in self.update_index().keys():
            del self[key]

    def __str__(self):
        return (
            f"{self.__class__.__name__} with keys {self.index.keys()}"
        )

    def __repr__(self):
        return self.__str__()


class Sticky:
    def __init__(
        self,
        address=None,
        value=None,
        cleanup_on_exit=False
    ):
        self.address = str(address)
        self.encode, self.decode = pickle_codec_factory()
        self._cached_value = value
        if cleanup_on_exit is True:
            atexit.register(self.close)

    @classmethod
    def note(
        cls,
        obj,
        address=None,
        # readonly=True,
        cleanup_on_exit=False
    ):
        init_kwargs = {
            'value': obj,
            'address': str(address),
            'cleanup_on_exit': cleanup_on_exit
        }
        if address is None:
            init_kwargs['address'] = f"{randint(100000, 999999)}"
        note = Sticky(**init_kwargs)
        note.stick(obj)
        return note

    def read(self, reread=False):
        if (self._cached_value is not None) and (reread is False):
            return self._cached_value
        try:
            block = SharedMemory(self.address)
        except FileNotFoundError:
            return None
        stream = block.buf.tobytes()
        self._cached_value = self.decode(stream)
        return self._cached_value

    def close(self):
        block = SharedMemory(self.address)
        block.unlink()
        block.close()

    def pull(self):
        return self.close()

    def stick(self, obj):
        encoded = self.encode(obj)
        size = len(encoded)
        block = open_block(self.address, size)
        block.buf[:] = encoded
        self._cached_value = obj

    def __str__(self):
        if self._cached_value is not None:
            return self._cached_value.__str__()
        else:
            return f'unread sticky'

    def __repr__(self):
        if self._cached_value is not None:
            return f"sticky: {self._cached_value.__repr__()}"
        else:
            return 'unread sticky'



"""functional utilities and generators"""
from functools import wraps, partial, reduce
from itertools import accumulate, repeat
from operator import add, contains, and_
from typing import Callable, Iterable, Any, Sequence, Collection


def pass_parameters(func, *args, **kwargs):
    return func(*args, **kwargs)


def catch_interaction(
    noninteractive: Any, func: Callable, *args, default: Any = "", **kwargs
):
    """
    if noninteractive is truthy, always return default. intended
    primarily as a wrapper to preempt attempts to prompt user input.
    """
    if noninteractive:
        return default
    return func(*args, **kwargs)


def naturals() -> accumulate:
    return accumulate(repeat(1), add)


def zero(*_, **__) -> None:
    """take anything, return nothing"""
    return


def triggerize(niladic_function: Callable) -> Callable:
    """
    implicitly turns a function into a trigger step for
    a pipeline
    """

    @wraps(niladic_function)
    def trigger(deferred_state=None):
        niladic_function()
        return deferred_state

    return trigger


def are_in(items: Collection, oper: Callable = and_) -> Callable:
    """
    iterable -> function
    returns function that checks if its single argument contains all
    (or by changing oper, perhaps any) items
    """

    def in_it(container: Collection) -> bool:
        inclusion = partial(contains, container)
        return reduce(oper, map(inclusion, items))

    return in_it


def is_it(*types: type) -> Callable[[Any], bool]:
    """partially-evaluated predicate form of `isinstance`"""

    def it_is(whatever: Any):
        return isinstance(whatever, types)

    return it_is


def intersection(*iterables: Iterable) -> set:
    return reduce(and_, [set(iterable) for iterable in iterables])


def disjoint(*sets: Iterable) -> list[list]:
    shared = intersection(*sets)
    return [
        [file for file in this_set if file not in shared] for this_set in sets
    ]


def constant(value: Any) -> Callable:
    def return_constant(*_, **__):
        return value

    return return_constant


def splat(func: Callable) -> Callable[[Sequence], Any]:
    @wraps(func)
    def splatified(args: Sequence) -> Any:
        return func(*args)

    return splatified



from operator import attrgetter
from typing import Optional, Union, Mapping, Any, Callable, Sequence, Hashable

from cytoolz import identity, keyfilter, first

from dustgoggles.func import naturals
from dustgoggles.structures import enumerate_as_mapping, reindex_mapping


class Composition:
    """
    class defining a composition of steps, optionally with
    additional input, output, and i/o points ("inserts" and "sends") --
    conceptually like a signal processing chain but _not_ designed for
    real-time signal processing.
    """
    def __init__(
        self,
        steps: Optional[
            Union[Mapping[Any, Callable], Sequence[Callable]]
        ] = None,
        sends: Optional[Union[Mapping[Any, Sequence], Sequence]] = None,
        inserts: Optional[
            Union[Mapping[Any, Mapping], Sequence[Mapping]]
        ] = None
    ):
        if steps is None:
            steps = [identity]
        self.steps = enumerate_as_mapping(steps)
        # sends and inserts could be defaultdicts, but in this case the
        # representation is ugly and the convenience is small.
        self.sends = enumerate_as_mapping(sends)
        self.inserts = enumerate_as_mapping(inserts)

    def _check_for_step(self, step_name: Hashable) -> Any:
        if step_name not in self.steps.keys():
            raise KeyError(f"{step_name} is not an element of the pipeline.")
        return self.steps[step_name]

    def add_step(self, step, name=None):
        """
        add step with name "name" -- by default, length of the pipeline + 1.
        if no step with that name currently exists, add it to the end of the
        pipeline; otherwise, replace the current step at "name".
        """
        if name is None:
            name = len(self.steps) + 1
        self.steps[name] = step

    def reindex(self):
        self.steps = reindex_mapping(self.steps)

    @property
    def index(self):
        return tuple(self.steps.keys())

    @property
    def insert_index(self):
        return tuple(self.inserts.keys())

    @property
    def send_index(self):
        return tuple(self.sends.keys())

    @property
    def function_names(self):
        return list(map(attrgetter("__name__"), self.steps.values()))

    def _perform_sends(self, step_name: Hashable, state: Any):
        """
        send pipeline state to send targets, if one matching this step name
        exists. unlike inserts, there is currently no provision for sends with
        value None; there probably should not be.
        """
        buses = self.sends.get(step_name)
        if (buses is None) or (len(buses) == 0):
            return
        for pipe, target, index in buses:
            if target is None:
                if pipe is None:
                    continue
                pipe(state)
            elif pipe is None:
                self.place_into(state, target, index)
            else:
                self.place_into(pipe(state), target, index)

    def add_send(
        self,
        step_name: Hashable,
        pipe: Optional[Callable] = None,
        target: Optional[Any] = None,
        pointer: Union[str, int, None] = None
    ):
        """
        adds a send to the pipeline after step_name; sends to pointer
        at target after processing through pipe. a string or int isas an
        insert into the step named 'target' of self. if no target is given,
        merely calls pipe and sends its output to nowhere. you are allowed
        to create sends from nonexistent step names, but they will do nothing
        unless that step name is later assigned.
        """
        if (target is None) and (pipe is None):
            raise ValueError(
                "Either pipe or target must be defined to create a send.")
        if step_name not in self.sends.keys():
            self.sends[step_name] = []
        self.sends[step_name].append((pipe, target, pointer))

    def add_insert(
        self,
        step_name: Hashable,
        key: Union[str, int],
        value: Any
    ):
        """
        adds an insert to the pipeline at step_name. an integer-valued
        "key" will turn into a positional argument to step; a string-valued
        "key" will turn into a keyword argument to the step at step name.
        "value" is simply the value of that arg or kwarg. you are allowed to
        add inserts to nonexistent steps, but they will do nothing unless that
        step name is later assigned.
        """
        if step_name not in self.inserts.keys():
            self.inserts[step_name] = {}
        self.inserts[step_name][key] = value

    # TODO: evaluate whether it might be important to base this tree on checks
    #  against the ABCs, which are of course unfortunately expensive -- but
    #  only trivally if this is being checked less than ~30 times per second.
    #  perhaps find some way to modify it at runtime.
    def place_into(self, thing, target, pointer):
        if isinstance(target, (str, int)):
            self.add_insert(target, pointer, thing)
        elif isinstance(target, list):
            if pointer is None:
                target.append(thing)
            else:
                target[pointer] = thing
        elif isinstance(target, dict):
            target[pointer] = thing
        elif "write" in dir(target):
            target.write(thing)

    @staticmethod
    def _flow_signal(state, args):
        # signal flows into the first open position, permitting partial
        # evaluation of steps with 'fixed' positional-only arguments
        opening = first(
            filter(lambda integer: integer - 1 not in args.keys(), naturals())
        )
        args[opening - 1] = state
        return [args[k] for k in sorted(args.keys())]

    def _call_partial_step(self, state, step, args, kwargs):
        # note that inserts always override args and kwargs baked into the
        # step itself.
        if "args" in dir(step):
            args = enumerate_as_mapping(step.args) | args
        if "keywords" in dir(step):
            kwargs = step.keywords | kwargs
        if len(args) == 0:
            return step.func(state, **kwargs)
        return step.func(*self._flow_signal(state, args), **kwargs)

    def call_step(self, step_name, state):
        step, insert = self.steps[step_name], self.inserts.get(step_name)
        if insert is None:
            args, kwargs = {}, {}
        else:
            args = keyfilter(lambda k: isinstance(k, int), insert)
            kwargs = keyfilter(lambda k: isinstance(k, str), insert)
        if "func" in dir(step):
            return self._call_partial_step(state, step, args, kwargs)
        if len(args) == 0:
            return step(state, **kwargs)
        return step(*self._flow_signal(state, args), **kwargs)

    def _do_step(self, step_name, state):
        """perform an individual step of the pipeline"""
        state = self.call_step(step_name, state)
        self._perform_sends(step_name, state)
        return state

    def itercall(self, signal: Any = None, **special_kwargs):
        self._bind_special_runtime_kwargs(special_kwargs)
        state = signal
        for step_name in self.steps.keys():
            state = self._do_step(step_name, state)
            yield state

    def execute(self, signal: Any = None, **special_kwargs):
        """execute the pipeline, initializing it with signal."""
        iterpipe = self.itercall(signal, **special_kwargs)
        state = None
        for state in iterpipe:
            pass
        return state

    def iter(self):
        raise NotImplementedError

    def _bind_special_runtime_kwargs(self, special_kwargs):
        """
        apply kwargs at time of execution across multiple steps
        according to domain-specific business logic.
        this exists for use by subclasses. it does nothing
        in the base Composition class.
        """
        pass

    def __str__(self):
        me_string = ""
        for attr in ("steps", "sends", "inserts"):
            if attr not in dir(self):
                continue
            me_string += f"{attr}:\n{getattr(self, attr).__repr__()}\n"
        if me_string == "":
            return "empty Composition"
        return me_string

    def __repr__(self):
        return self.__str__()



import random
import string
from multiprocessing import Pipe, Process
from typing import Hashable, Callable

from cytoolz import get_in

from dustgoggles.structures import NestingDict

RNG = random.Random()


# TODO, maybe: make this a hypothesis strategy?
def randval(itemtype):
    if itemtype is str:
        return "".join(RNG.choices(string.ascii_letters, k=RNG.randint(1, 10)))
    elif itemtype is int:
        return RNG.randint(-100000, 100000)
    elif itemtype is float:
        return round((RNG.random() - 0.5) * 200000, 2)
    elif itemtype is tuple:
        return tuple(
            round((RNG.random() - 0.5) * 200000, 2)
            for _ in range(RNG.randint(1, 10))
        )
    elif isinstance(itemtype, Callable):
        return itemtype()
    raise TypeError(f"no generator for {itemtype}")


def pick_key_at_level(mapping, keys, hashables, branch_weight):
    level_mapping = get_in(keys, mapping)
    nestkeys = [
        k for k, v in level_mapping.items() if isinstance(v, NestingDict)
    ]
    if (len(nestkeys) == 0) or (RNG.random() > branch_weight):
        nestkey = randval(RNG.choice(hashables))
    else:
        nestkey = RNG.choice(nestkeys)
    return nestkey


def random_nested_dict(
    quantity,
    maxdepth=4,
    types=(str, int, float, tuple),
    keytypes=None,
    branch_weight=0.4
):
    nest = NestingDict()
    levels = tuple(range(maxdepth))
    if keytypes is None:
        keytypes = [t for t in types if isinstance(t, Hashable)]
    for _ in range(quantity):
        level = RNG.choice(levels)
        keys = []
        for level_ix in range(level):
            keys.append(
                pick_key_at_level(nest, keys, keytypes, branch_weight)
            )
        get_in(keys, nest)[
            randval(RNG.choice(keytypes))
        ] = randval(RNG.choice(types))
    return nest.todict()


def piped(func):
    here, there = Pipe()

    def sendback(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except Exception as ex:
            result = ex
        return there.send(result)

    return here, sendback


def wrap_piped(func, block=True):
    def get_results_from_child_process(*args, **kwargs):
        here, sendback = piped(func)
        proc = Process(target=sendback, args=args, kwargs=kwargs)
        proc.start()
        if block is True:
            proc.join()
            result = here.recv()
            proc.close()
            return result
        return proc

    return get_results_from_child_process



# import sys
# from multiprocessing.shared_memory import SharedMemory
#
# import numpy as np

# from dustgoggles.structures import dig_and_edit


# def reference_shared_memory_arrays(
#     block_info, fetch=True
# ) -> tuple[dict, dict]:
#     blocks = {
#         variable: SharedMemory(name=info["name"])
#         for variable, info in block_info.items()
#     }
#     if fetch is True:
#         chunk = {
#             variable: np.ndarray(
#                 info["shape"], dtype=info["dtype"], buffer=blocks[variable].buf
#             )
#             for variable, info in block_info.items()
#         }
#     else:
#         chunk = {}
#     return blocks, chunk

#
#
# def share_mapping(mapping, codec=share_array):
#     block_info = {}
#
#     for key, value in mapping.items():
#         block_info[key] =
#     return block_info

#
# def slice_into_shared_chunks(chunksz, data, nphots):
#     variable_names = [key for key in data.keys()]
#     chunk_slices = make_chunk_slices(chunksz, nphots)
#     total_chunks = len(chunk_slices)
#     block_directory = {}
#     for chunk_ix in range(total_chunks):
#         block_directory = slice_chunk_into_memory(
#             block_directory, chunk_ix, data, chunk_slices, variable_names
#         )
#     return block_directory
#

# def make_chunk_slices(chunksz, nphots):
#     table_indices = []
#     total_chunks = range(int(nphots / chunksz) + 1)
#     for chunk_ix in total_chunks:
#         chunkbeg, chunkend = chunk_ix * chunksz, (chunk_ix + 1) * chunksz
#         if chunkend > nphots:
#             chunkend = None
#         table_indices.append((chunkbeg, chunkend))
#     return table_indices
#
#
# def slice_into_memory(data, indices):
#     return send_to_shared_memory(
#         {key: value[slice(*indices)] for key, value in data.items()}
#     )


# def slice_chunk_into_memory(
#     block_directory, chunk_ix, data, table_indices, variable_names=None
# ):
#     arrays = [
#         array[slice(*table_indices[chunk_ix])] for array in data.values()
#     ]
#     if variable_names is None:
#         variable_names = tuple(data.keys())
#     block_info = send_to_shared_memory(
#         {
#             variable_name: array
#             for variable_name, array in zip(variable_names, arrays)
#         }
#     )
#     block_directory[chunk_ix] = block_info
#     return block_directory

#
# def send_mapping_to_shared_memory(
#     mapping,
#     filter_func=None,
#     setter_func=lambda k, v: memorize_array(v)
# ):
#     block_directory = dig_and_edit(
#         mapping, filter_func, setter_func
#     )
#     return block_directory
#

# def unlink_nested_block_dict(cal_data):
#     all_cal_blocks = []
#     for cal_name, cal_info in cal_data.items():
#         if cal_info is None:
#             continue
#         cal_blocks, _ = reference_shared_memory_arrays(cal_info, fetch=False)
#         all_cal_blocks += list(cal_blocks.values())
#     for block in all_cal_blocks:
#         block.close()
#         block.unlink()
#
#
# def get_column_from_shared_memory(results, column_name, unlink=True):
#     column_info = {
#         chunk_ix: results[chunk_ix][column_name]
#         for chunk_ix in sorted(results.keys())
#     }
#     blocks, column_slices = reference_shared_memory_arrays(column_info)
#     column = np.hstack(list(column_slices.values()))
#     del column_slices
#     if unlink is True:
#         for block in blocks.values():
#             block.close()
#             block.unlink()
#     return column
#



from setuptools import setup, find_packages

setup(
    name="dustgoggles",
    version="0.5.0",
    url="https://github.com/millionconcepts/dustgoggles.git",
    author="Million Concepts",
    author_email="mstclair@millionconcepts.com",
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=["cytoolz"],
    extras_require={
        "pivot": ["numpy", "pandas"],
        "tests": ["pytest"],
        "array_codecs": ["numpy"]
    }
)



from multiprocessing.shared_memory import SharedMemory
import sys


def open_block(address, size, create=True, overwrite=False):
    """
    open and return a shared memory block with the specified size. if
    `create` is True, create a new block if none exists. if `overwrite` is
    True, also unlink/overwrite any existing block at that address.
    """
    try:
        return SharedMemory(address, size=size, create=create)
    except FileExistsError:
        if overwrite is False:
            return SharedMemory(address)
        old_block = SharedMemory(address)
        old_block.unlink()
        old_block.close()
        return SharedMemory(address, True, size)


def fetch_block_bytes(address: str) -> bytes:
    """
    get the contents of the shared memory block at address as a binary blob.
    """
    block = SharedMemory(address)
    return block.buf.tobytes()


def exists_block(address: str) -> bool:
    """
    return True if a shared memory block exists at address; False if not.
    """
    try:
        SharedMemory(address)
        return True
    except FileNotFoundError:
        return False
    except ValueError as error:
        if "cannot mmap an empty file" in str(error):
            return True
        raise



def delete_block(address: str, missing_ok=False):
    """delete the shared memory block at address."""
    try:
        block = SharedMemory(address)
        block.unlink()
        block.close()
    except FileExistsError:
        if missing_ok is False:
            raise


def deactivate_shared_memory_resource_tracker():
    """
    monkeypatch resource tracker on posix systems to handle longstanding
    excessively-enthusiastic shared-memory gc issue. calling this function
    from a process will prevent the Python resource tracker from
    "cleaning up" any shared memory blocks in use by that process when the
    process terminates. by default, on Mac and Linux, the tracker does this
    _whether or not_ blocks are also in use by other processes, except in
    cases in which all processes using the blocks descend from the same
    parent process. this is the equivalent of not being able to open a file
    without deleting it; sometimes it doesn't matter, sometimes it's
    catastrophic.

    see: https://github.com/python/cpython/issues/82300 (also direct source
    for this code)

    note that this _will_ permit memory leaks from non-gracefully-terminated
    processes if you call it with excessive enthusiasm and aren't tracking
    the shared memory blocks anywhere else.
    """
    if sys.platform in ('win32', 'cygwin'):
        return

    # noinspection PyUnresolvedReferences
    from multiprocessing import resource_tracker

    # noinspection PyUnresolvedReferences
    def fix_register(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.register(self, name, rtype)
    resource_tracker.register = fix_register

    # noinspection PyUnresolvedReferences
    def fix_unregister(name, rtype):
        if rtype == "shared_memory":
            return
        return resource_tracker._resource_tracker.unregister(self, name, rtype)
    resource_tracker.unregister = fix_unregister

    if "shared_memory" in resource_tracker._CLEANUP_FUNCS:
        del resource_tracker._CLEANUP_FUNCS["shared_memory"]


